//! A container representing a moment in ISO-8601 format Time
use std::{fmt, str::FromStr};

mod error;
mod parser;
pub use {error::Error, parser::Parser};

#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
pub enum Sign {
    Positive,
    Negative,
}

impl fmt::Display for Sign {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Positive => write!(f, "+"),
            Self::Negative => write!(f, "-"),
        }
    }
}

impl TryFrom<char> for Sign {
    type Error = Error;

    fn try_from(value: char) -> Result<Self, Self::Error> {
        match value {
            '+' => Ok(Self::Positive),
            '-' => Ok(Self::Negative),
            _ => Err(Error::InvalidOffset),
        }
    }
}

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
pub struct Offset {
    pub sign: Sign,
    pub hours: u8,
    pub minutes: Option<u8>,
}

impl fmt::Display for Offset {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(m) = self.minutes {
            write!(f, "{}{:02}:{m:02}", self.sign, self.hours)
        } else {
            write!(f, "{}{:02}", self.sign, self.hours)
        }
    }
}

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
pub enum TimeZone {
    UTC,
    Offset(Offset),
}

impl fmt::Display for TimeZone {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::UTC => write!(f, "Z"),
            Self::Offset(o) => write!(f, "{o}"),
        }
    }
}

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
pub struct DateTime {
    pub year: u32,
    pub month: Option<u8>,
    pub day: Option<u8>,
    pub hour: Option<u8>,
    pub minute: Option<u8>,
    pub second: Option<u8>,
    pub tz: Option<TimeZone>,
}

impl fmt::Display for DateTime {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:04}", self.year)?;
        'blk: {
            for field in &[self.month, self.day] {
                if let Some(field) = field {
                    write!(f, "-{field:02}")?;
                } else {
                    break 'blk;
                }
            }
            if let Some(h) = self.hour {
                write!(f, "T{h:02}")?;
            } else {
                break 'blk;
            }
            for field in &[self.minute, self.second] {
                if let Some(field) = field {
                    write!(f, ":{field:02}")?;
                } else {
                    break 'blk;
                }
            }
        }
        if let Some(ref tz) = self.tz {
            write!(f, "{tz}")
        } else {
            Ok(())
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn fmt_full() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: Some(3),
            second: Some(8),
            tz: Some(TimeZone::Offset(Offset {
                sign: Sign::Positive,
                hours: 5,
                minutes: Some(15),
            })),
        };
        assert_eq!(dt.to_string(), "2023-05-07T09:03:08+05:15");
    }

    #[test]
    fn fmt_partial_offset() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: Some(3),
            second: Some(8),
            tz: Some(TimeZone::Offset(Offset {
                sign: Sign::Negative,
                hours: 5,
                minutes: None,
            })),
        };
        assert_eq!(dt.to_string(), "2023-05-07T09:03:08-05");
    }

    #[test]
    fn fmt_utc() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: Some(3),
            second: Some(8),
            tz: Some(TimeZone::UTC),
        };
        assert_eq!(dt.to_string(), "2023-05-07T09:03:08Z");
    }

    #[test]
    fn fmt_no_tz() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: Some(3),
            second: Some(8),
            tz: None,
        };
        assert_eq!(dt.to_string(), "2023-05-07T09:03:08");
    }

    #[test]
    fn fmt_no_seconds() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: Some(3),
            second: None,
            tz: Some(TimeZone::UTC),
        };
        assert_eq!(dt.to_string(), "2023-05-07T09:03Z");
    }

    #[test]
    fn fmt_no_minutes() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: Some(9),
            minute: None,
            second: Some(50),
            tz: Some(TimeZone::UTC),
        };
        assert_eq!(dt.to_string(), "2023-05-07T09Z");
    }

    #[test]
    fn fmt_no_time() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: Some(7),
            hour: None,
            minute: None,
            second: Some(50),
            tz: Some(TimeZone::UTC),
        };
        assert_eq!(dt.to_string(), "2023-05-07Z");
    }

    #[test]
    fn fmt_no_day() {
        let dt = DateTime {
            year: 2023,
            month: Some(5),
            day: None,
            hour: Some(20),
            minute: None,
            second: Some(50),
            tz: None,
        };
        assert_eq!(dt.to_string(), "2023-05");
    }

    #[test]
    fn fmt_no_month() {
        let dt = DateTime {
            year: 2023,
            month: None,
            day: None,
            hour: Some(20),
            minute: None,
            second: Some(50),
            tz: None,
        };
        assert_eq!(dt.to_string(), "2023");
    }
}
