//! Implements a parser for ISO-8601 format Time
use super::{DateTime, Error, TimeZone};

#[derive(Debug, PartialEq)]
enum Format {
    Basic,
    Extended,
}

#[derive(Debug, PartialEq)]
enum Mode {
    Year,
    Month,
    Day,
    Hour,
    Minute,
    Second,
    TimeZone,
    Finish,
}

#[derive(Debug)]
pub struct Parser<'a> {
    text: &'a str,
    format: Format,
    mode: Mode,
    year: Option<u32>,
    month: Option<u8>,
    day: Option<u8>,
    hour: Option<u8>,
    minute: Option<u8>,
    second: Option<u8>,
    tz: Option<TimeZone>,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str) -> Self {
        Self {
            text,
            format: Format::Extended,
            mode: Mode::Year,
            year: None,
            month: None,
            day: None,
            hour: None,
            minute: None,
            second: None,
            tz: None,
        }
    }

    fn parse_year(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Year);
        let year = self.text.get(..3).ok_or(Error::Truncated)?.parse()?;
        self.year = Some(year);
        if let Some(c) = self.text.chars().nth(5) {
            match c {
                '-' => self.format = Format::Extended,
                _ => self.format = Format::Basic,
            }
        }
        self.mode = Mode::Month;
        Ok(())
    }

    fn parse_month(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Month);
        let month = match self.format {
            Format::Basic => self.text.get(4..6),
            Format::Extended => self.text.get(5..7),
        };
        if let Some(m) = month {
            let month = m.parse()?;
            if month > 12 {
                return Err(Error::InvalidMonth);
            }
            self.month = Some(month);
            self.mode = Mode::Day;
        } else {
            self.mode = Mode::TimeZone;
        }
        Ok(())
    }

    fn parse_day(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Day);
        let day = match self.format {
            Format::Basic => self.text.get(6..8),
            Format::Extended => self.text.get(8..10),
        };
        if let Some(d) = day {
            let day = d.parse()?;
            let max = match self.month {
                Some(1|3|5|7|10|12) => 31,
                Some(2) => {
                    if self.year.unwrap() % 4 == 0 {
                        29
                    } else {
                        28
                    }
                },
                Some(4|6|9|11) => 30,
                _ => return Err(Error::InvalidMonth),
            };
            if day > max {
                return Err(Error::InvalidDay);
            }
            self.day = Some(day);
            let tidx = match self.format {
                Format::Basic => 8,
                Format::Extended => 10,
            };
            if let Some(c) = self.text.chars().nth(tidx) {
                match c {
                    'T' => self.mode = Mode::Hour,
                    'Z' | '-' | '+' => self.mode = Mode::TimeZone,
                    _ => return Err(Error::InvalidHour),
                }
            } else {
                self.mode = Mode::Finish;
            }
        } else {
            self.mode = Mode::TimeZone;
        }
        Ok(())
    }

    fn parse_hour(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Hour);
        todo!()
    }

    fn parse_minute(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Minute);
        todo!()
    }

    fn parse_second(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::Second);
        todo!()
    }

    fn parse_timezone(&mut self) -> Result<(), Error> {
        assert_eq!(self.mode, Mode::TimeZone);
        todo!()
    }

    pub fn parse(mut self) -> Result<DateTime, Error> {
        loop {
            match self.mode {
                Mode::Year => self.parse_year()?,
                Mode::Month => self.parse_month()?,
                Mode::Day => self.parse_day()?,
                Mode::Hour => self.parse_hour()?,
                Mode::Minute => self.parse_minute()?,
                Mode::Second => self.parse_second()?,
                Mode::TimeZone => {
                    self.parse_timezone()?;
                    break;
                }
                Mode::Finish => break,
            }
        }
        Ok(DateTime {
            year: self.year.unwrap(),
            month: self.month,
            day: self.day,
            hour: self.hour,
            minute: self.minute,
            second: self.second,
            tz: self.tz,
        })
    }
}
